from __future__ import absolute_import
import argparse
from collections import OrderedDict

import logging
import os
import pkg_resources
import re
import socket
import subprocess
import sys
import urllib

from conversion.boolean import convert_bool
import requests

from .config import get_config
from .websocket import websocket
from zymbit.config import get_config


# TODO: look into flankr instead of this
# http://blog.mailgun.com/we-just-open-sourced-flanker-our-python-email-address-and-mime-parsing-library/
EMAIL_ADDRESS_RE = re.compile(r'.+@.+\..+')


def find_cpuinfo(markers):
    # no support for non-linux systems
    if os.path.exists('/proc/cpuinfo'):
        content = open('/proc/cpuinfo', 'rb').read()
        for marker, return_value in markers.items():
            if marker in content:
                return return_value


def get_connection(url=None, credentials=None, check_hostname=None):
    headers = []

    logger = logging.getLogger(__name__)

    _config = get_config()

    check_hostname = check_hostname or _config.lookup('cloud.check_hostname')

    sslopt = {"check_hostname": check_hostname}

    url = url or _config.lookup('cloud.websocket_url')
    request_url = url
    if credentials:
        connector_version = pkg_resources.require('zymbit')[0].version

        credentials.update({
            'hostname': socket.gethostname(),
            'connector_version': connector_version,
        })

        credentials.update(get_device_meta())

        encoded = urllib.urlencode(credentials)
        request_url = '{}?{}'.format(url, encoded)

    logger.debug('getting websocket connection at url={}'.format(url))

    ws = websocket.create_connection(request_url, header=headers, sslopt=sslopt)
    ws.settimeout(0)

    return ws


def get_device_meta():
    device_meta = {}

    try:
        # TODO: support linux
        # this didn't work on a server container, so just removed this altogether
        # because these params are intended for additional device metadata, but it
        # would be great to be able to get this from any system.
        from .compat import util
    except ImportError:
        pass
    else:
        device_meta.update({
            'distro': util.get_distro(),
            'model': util.get_model(),
            'system': util.get_system(),
        })

    return device_meta


def get_model():
    systems = OrderedDict((
        ('Arduino Yun', 'yun'),
        ('BCM2708', '1'),
        ('BCM2709', '2'),
    ))
    cpuinfo = find_cpuinfo(systems)
    if cpuinfo:
        return cpuinfo

    if sys.platform == 'darwin':
        return sys.platform


def get_provisioning_token(config=None):
    api = ZymbitApi(config=config)
    response = api.request('get', '/organizations')

    data = response.json()
    organization = data['results'][0]
    provisioning_group = organization['provisioning_groups'][0]

    return provisioning_group['token']


def get_system():
    systems = OrderedDict((
        ('Arduino Yun', 'arduino'),
        ('BCM270', 'raspberrypi'),  # note this will match BCM2708 (rpi) and BCM2709 (rpi2)
    ))
    cpuinfo = find_cpuinfo(systems)
    if cpuinfo:
        return cpuinfo

    if sys.platform == 'darwin':
        return sys.platform

    if 'linux' in sys.platform:
        return 'linux'


def is_interactive():
    import __main__ as main
    return hasattr(main, '__file__')


def run_command(command):
    return subprocess.Popen(command).wait()


def validate_email(address):
    """
    verify the email address given is valid
    :param address: string - an email address
    :return: address or raise an exception
    """
    matches = EMAIL_ADDRESS_RE.match(address)
    if matches is None:
        raise argparse.ArgumentTypeError('address={} does not appear to be valid'.format(address))

    return address


class ZymbitApi(object):
    def __init__(self, config=None):
        self.response = None
        self._config = config

    @property
    def config(self):
        return self._config or get_config()

    def request(self, method, endpoint, data=None):
        config = self.config

        api_url = config['cloud']['api_url']
        verify = config['cloud'].get('check_hostname', True)

        default_auth = config['auth']['default_auth']
        password = config['auth'][default_auth]['password']
        auth = (default_auth, password)

        url = '{}{}'.format(api_url, endpoint)

        method_fn = getattr(requests, method)

        try:
            self.response = method_fn(url, auth=auth, data=data, verify=verify)
        except requests.ConnectionError, exc:
            raise requests.ConnectionError('Unable to connect to url={}'.format(url))
        else:
            self.response.raise_for_status()

        return self.response



from __future__ import absolute_import

import Queue
import logging
import os
import random
import select
import socket
import ssl
import time
import datetime

import requests

from .exceptions import NotConnected, ProvisionError
from .websocket import WebSocketConnectionClosedException
from zymbit.clientid import get_client_id

from zymbit.compat import util, config
from zymbit.util import get_connection, get_device_meta
from zymbit.envelope import get_envelope
from zymbit.timeutil import now

MAX_BACKOFF = 300  # 5 minutes
BACKOFF_JITTER = 10  # 10 seconds
CONNECTION_SETTLE_DT = datetime.timedelta(seconds=10)  # 10 second minimum to clear backoff

PROVISION_PATH = os.environ.get('PROVISION_PATH', '/provision')


def get_logger():
    return logging.getLogger(__name__)


def register(provisioning_token=None):
    logger = get_logger()

    client_id = get_client_id()
    _config = config.get_config()

    provisioning_token = provisioning_token or config.get_config().lookup('auth.provisioning_token')
    if not provisioning_token:
        raise ProvisionError('Provisioning token not found')

    logger.debug('provisioning, client_id={}'.format(client_id))

    data = {
        'token': provisioning_token,
        'client_id': client_id,
    }

    data.update(get_device_meta())

    # TODO: update to use ZymbitApi object
    provision_url = '{}{}'.format(_config.lookup('cloud.api_url'), config.PROVISION_PATH)
    verify = _config.lookup('cloud.check_hostname')
    try:
        response = requests.post(provision_url, data=data, verify=verify)
    except requests.ConnectionError, exc:
        logger.exception(exc)
        logger.error('error making request to provision_url={}'.format(provision_url))

        raise

    try:
        response.raise_for_status()
    except requests.HTTPError, exc:
        logger.exception(exc)
        logger.error('error making request to provision_url={}, content={}'.format(
            provision_url, response.content)
        )

        raise

    response_data = response.json()

    return response_data


class Client(object):
    def __init__(self, handle_disconnect=True, connected_callback=None, client_id=None):
        """
        :param handle_disconnect - bool: internally handle a disconnect an gracefully reconnect; raises exception when False
        :return:
        """
        self.client_id = client_id

        self._connected_callback = connected_callback

        self.handle_disconnect = handle_disconnect

        self._ws = None
        self.is_connected = False
        self.connected_at = None

        self.backoff = 0
        self.backoff_until = None

        self.recv_q = Queue.Queue()

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @property
    def check_hostname(self):
        _config = config.get_config()

        return _config.get('cloud', {}).get('check_hostname')

    def connect(self):
        # simply call the ws property that takes care of establishing a connection
        return self.ws

    @property
    def credentials(self):
        env_uuid = os.environ.get('ZYMBIT_AUTH_UUID')
        env_token = os.environ.get('ZYMBIT_AUTH_TOKEN')

        if env_uuid and env_token:
            auth = {
                'uuid': env_uuid,
                'auth_token': env_token,
            }
        else:
            _config = config.get_config()
            client_id = get_client_id()

            auth = _config.get('auth', {}).get(client_id)

        return auth

    def _clear_backoff(self):
        if self.connected_at is None:
            return

        # only clear backoff when the connection has been established for more than CONNECTION_SETTLE
        if now() > self.connected_at + CONNECTION_SETTLE_DT:
            self.logger.debug(
                'connected more than {} clearing backoff'.format(CONNECTION_SETTLE_DT)
            )

            self.connected_at = None
            self.backoff = 0
            self.backoff_until = None

            return True

    def close(self):
        # don't close on closed
        if self._ws is None:
            return

        # self._ws.close()
        self.is_connected = False
        self._ws = None

        cleared = self._clear_backoff()
        if not cleared:
            self.backoff += 1
            seconds = min(2**self.backoff, MAX_BACKOFF)
            jitter = datetime.timedelta(seconds=BACKOFF_JITTER*random.random())
            self.backoff_until = jitter + now() + \
                                 datetime.timedelta(seconds=seconds)

            self.logger.debug('backoff={}, backoff_until={}, jitter={}'.format(self.backoff, self.backoff_until, jitter))

    def connected(self):
        """
        Called when websocket connection is initialized
        """
        self.is_connected = True

        self.connected_at = now()
        if self._connected_callback:
            self._connected_callback()

    def fileno(self):
        return self.ws.sock.fileno()

    def recv(self, ws=None):
        ws = ws or self.ws
        sock = ws.sock

        # check if there are pending messages in the queue
        try:
            return self.recv_q.get_nowait()
        except Queue.Empty:
            pass

        r, _, _ = select.select([sock], [], [], 0.01)
        if sock not in r:
            return

        try:
            return ws.recv()
        except WebSocketConnectionClosedException, exc:
            self.close()

            if not self.handle_disconnect:
                raise
        except NotConnected:
            if not self.handle_disconnect:
                raise
        except ssl.SSLError, exc:
            # SSLError: [Errno 8] _ssl.c:1359: EOF occurred in violation of protocol -- arduino
            error_number = exc[0]
            if error_number not in (2, 8):
                self.logger.warning('ssl error #{}, closing connection'.format(error_number))

            self.close()
            if not self.handle_disconnect:
                raise
        except socket.error as exc:
            error_number = exc[0]
            if error_number not in (11, 54):
                self.logger.warning('got socket error {}'.format(error_number))

            self.close()
            if not self.handle_disconnect:
                raise

    def send(self, action, params=None):
        """
        Wrap the given action and params in an envelope and send

        This calls the get_envelope() utility prior to sending the given message.

        :param action: string - upstream action
        :param params: dictionary - contains any data for this action

        :return: the result of send_raw()
        """
        return self.send_raw(get_envelope(action, params=params, client_id=self.client_id))

    def send_raw(self, message):
        """
        Send raw message

        Do not wrap in an envelope; simply send what was passed in

        :param message: string
        :return: result of websocket.send()
        """
        try:
            return self.ws.send(message)
        except socket.error, exc:
            # error: [Errno 131] Connection reset by peer -- arduino
            error_number = exc[0]
            if error_number not in (32, 131):
                self.logger.exception(exc)
                self.logger.warning('socket error_number={}'.format(error_number))

            self.close()

            if not self.handle_disconnect:
                raise

            return False
        except WebSocketConnectionClosedException:
            self.close()

            if not self.handle_disconnect:
                raise

            return False  # the message was not sent

    @property
    def url(self):
        _config = config.get_config()

        return _config.lookup('cloud.websocket_url')

    @property
    def ws(self):
        if self.backoff_until and self.backoff_until > now():
            raise NotConnected()

        if self._ws is not None:
            return self._ws

        try:
            self._ws = get_connection(url=self.url, credentials=self.credentials, check_hostname=self.check_hostname)
        except socket.error, exc:
            # error: [Errno 146] Connection refused -- arduino yun
            # error: [Errno 145] Connection timed out -- arduino yun
            # error: [Errno 148] No route to host -- arduino yun
            # gaierror: [Errno -2] Name or service not known -- arduino yun
            # error: [Errno 131] Connection reset by peer -- arduino yun
            error_number = exc[0]
            self.logger.warning('unable to connect to url={}, socket.error number {}'.format(self.url, error_number))

            self.close()
            raise NotConnected(exc)

        for _ in range(1000):
            try:
                message = self.recv(ws=self._ws)  # prevent recv() from calling this property
            except Exception, exc:
                raise

            # self.recv() may close the connection, so check if _ws is not None
            if self._ws is None or not self._ws.connected:
                self.close()
                raise NotConnected('connection closed on opening')

            if message:
                self.recv_q.put_nowait(message)

                self.logger.debug('message={}'.format(message))
                self.connected()

                break

            time.sleep(0.01)
        else:
            self.close()

            raise NotConnected('unable to get websocket connection')

        self.logger.debug('got websocket, _ws={}'.format(self._ws))

        return self._ws



import logging

from tabulate import tabulate

from zymbit.commands import Command

from zymbit.util import ZymbitApi


class StreamsCommand(Command):
    command_name = 'streams'
    datasets_endpoint = '/datasets'
    devices_endpoint = '/zymbots'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        api = ZymbitApi()

        # get devices
        response = api.request('get', self.devices_endpoint)

        devices_data = response.json()
        devices_by_uuid = dict([(x['uuid'], x) for x in devices_data['results']])

        response = api.request('get', self.datasets_endpoint)

        data = response.json()

        headers = ['zymbot', 'routing_key']
        table = [[devices_by_uuid.get(x['zymbot']).get('hostname'), x['path']]
                 for x in data['results']]
        table.sort(key=lambda item: '{}{}'.format(item[0], item[1]))

        print tabulate(table, headers=headers)



import json
import logging
import os
import requests
import stat

from conversion import convert_bool
from zymbit.clientid import get_client_id

AUTH_TOKEN_ENDPOINT = '/auth_token'
PROVISION_PATH = os.environ.get('PROVISION_PATH', '/provision')

TUNNEL_SSH_KEY_PATH = os.path.expanduser(os.environ.get('TUNNEL_SSH_KEY_PATH', '~/.ssh/id_tunnel'))

INIT_SCRIPT_PATH = '/etc/init.d/zymbit'
FILES_ROOT = None

LOGFILE = '/var/log/zymbit.log'

# setup a chain of configurations to try
CONFIG_PATH = os.environ.get('CONFIG_PATH', '/etc/zymbit.conf')
CONFIG_PATHS = (
    CONFIG_PATH,
    os.path.expanduser('~/.zymbit.conf'),
)


def get_logger():
    return logging.getLogger(__name__)


class ConfigError(Exception):
    pass


class ConfigDict(dict):
    def lookup(self, path):
        section = self

        for item in path.split('.'):
            section = section.get(item)
            if section is None:
                break

        # return True/False as is
        if isinstance(section, bool):
            return section

        return section


def get_auth_token(username, password):
    config = get_config()

    api_url = config.lookup('cloud.api_url')
    verify = config.lookup('cloud.check_hostname')
    url = '{}{}'.format(api_url, AUTH_TOKEN_ENDPOINT)
    params = {
        'username': username,
        'password': password,
    }

    response = requests.post(url, data=params, verify=verify)
    response.raise_for_status()

    token_data = response.json()

    return token_data['token']


def get_config():
    logger = get_logger()

    config = ConfigDict({
        'cloud': {
            'api_url': os.environ.get('API_URL', 'https://zymbit.com/api/v1'),
            'check_hostname': convert_bool(os.environ.get('SSL_CHECK_HOSTNAME', 'true')),
            'websocket_url': os.environ.get('WEBSOCKET_URL', 'wss://ws.zymbit.com/channel'),
        }
    })

    for config_path in CONFIG_PATHS:
        if os.path.exists(config_path):
            try:
                _config = json.load(open(config_path, 'rb'))
            except ValueError:
                logger.warning('Unable to load config at {}'.format(config_path))
            else:
                config.update(_config)

    return config


def save_auth(auth_data, password):
    config = get_config()

    email = auth_data['email']
    auth_data['password'] = password

    auth_config = config.setdefault('auth', {})

    auth_config['default_auth'] = email
    auth_config[email] = auth_data

    # request an auth token
    try:
        auth_token = get_auth_token(email, password)
    except requests.HTTPError:
        raise
    else:
        auth_data['auth_token'] = auth_token

    # check for a device token
    client_id = get_client_id()
    if client_id not in auth_config:
        from zymbit.client import register
        from zymbit.util import get_provisioning_token

        provisioning_token = get_provisioning_token(config=config)

        device_config = register(provisioning_token=provisioning_token)
        auth_config[client_id] = device_config['auth'][client_id]

    save_config(config)


def save_config(config):
    logger = logging.getLogger(__name__)

    serialized = json.dumps(config)

    for config_path in CONFIG_PATHS:
        try:
            with open(config_path, 'wb') as fh:
                fh.write(serialized)

            os.chmod(config_path, stat.S_IRUSR | stat.S_IWUSR)
        except IOError, exc:
            # IOError: [Errno 13] Permission denied: '[...]'
            if exc[0] != 13:
                raise
        else:
            logger.debug('config saved at {}'.format(config_path))
            break
    else:
        raise ConfigError('unable to save configuration at any location: {}'.format(CONFIG_PATHS))



import os
import re

from subprocess import PIPE, Popen

CLIENT_ID = os.environ.get('CLIENT_ID')
CLIENT_ID_RE = re.compile(r'(?P<device>\w+).* HWaddr (?P<mac>[0-9a-fA-F:]+).*')
MAC_CLIENT_ID_RE = re.compile(r'(?P<device>\w+): .*')
MAC_MAC_RE = re.compile(r'\s+ether (?P<mac>[0-9a-fA-F:]+).*')

IGNORED_DEVICES = (
    'docker0',
)

def get_output(command):
    proc = Popen(command, stdout=PIPE)

    return proc.communicate()[0].splitlines()


def get_client_id():
    global CLIENT_ID

    if CLIENT_ID:
        return CLIENT_ID

    devices = {}

    _device_name = None

    lines = get_output('/sbin/ifconfig')
    for line in lines:
        matches = CLIENT_ID_RE.match(line)
        if matches:
            t = matches.groupdict()
            devices[t['device']] = t['mac']
            continue

        matches = MAC_CLIENT_ID_RE.match(line)
        if matches:
            _device_name = matches.group('device')
            continue

        matches = MAC_MAC_RE.match(line)
        if matches and _device_name and _device_name not in IGNORED_DEVICES:
            devices[_device_name] = matches.group('mac')
            _device_name = None

    if not devices:
        raise Exception('Unable to find client ID')

    if 'eth0' in devices:
        CLIENT_ID = devices['eth0']
    elif 'en0' in devices:
        CLIENT_ID = devices['en0']
    else:
        CLIENT_ID = devices.values()[0]

    return CLIENT_ID


if __name__ == '__main__':
    print get_client_id()



import datetime
import logging

# it's impossible that "now" is less than this datetime
# we know we are out of sync with real time if we ever
# get a time value less than this
MIN_DT = datetime.datetime(2014, 7, 25, 17, 00, 00)  # Zymbit est date, UTC

LONG_TIME_AGO = datetime.datetime(1, 1, 1)  # a really long time ago


def now():
    return datetime.datetime.utcnow()


def timestamp(dt=None):
    if dt is None:
        dt = now()

    return dt.isoformat('T')


def get_sleep_time(seconds, start):
    """
    Wait at most the given number of seconds from the initial time given
    :param seconds: float - number of seconds to wait
    :param start: datetime - the start time
    :return: float - time to wait
    """
    _now = now()
    delta = _now - start
    diff = delta.seconds + (1.0 * delta.microseconds / 1e6)
    wait = max(0, seconds - diff)

    # print 'start={}, _now={}, delta={}, diff={}, wait={}'.format(start, _now, delta, diff, wait)

    return wait


class MillisDatetime(object):
    def __init__(self, millis):
        self.last_millis = None
        self.initial = None

        self.set_initial(millis)

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def get_time(self, millis):
        if millis < self.last_millis:
            self.logger.info(
                'time rolled over, last_millis={}, millis={}'.format(
                    self.last_millis, millis))

            self.set_initial(millis)

        delta = datetime.timedelta(milliseconds=millis)
        return self.initial + delta

    def set_initial(self, millis):
        delta = datetime.timedelta(milliseconds=millis)
        self.initial = now() - delta

        self.last_millis = millis



import json
import logging
import socket

from select import select
from zymbit.connector import get_connector
from zymbit.envelope import parse_buf
from zymbit.exceptions import NotConnected

BUFSIZE = 4096


class BaseServer(object):
    def __init__(self, host, port, message_handler=None):
        self.addr = (host, port)

        self._tcp_sock = None
        self._udp_sock = None

        self.connections = {}

        self.message_handler = message_handler

        self._run = True

    @property
    def logger(self):
        logger_name = '{}.{}'.format(__name__, self.__class__.__name__)
        return logging.getLogger(logger_name)

    @property
    def tcp_sock(self):
        if self._tcp_sock:
            return self._tcp_sock

        try:
            self._tcp_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self._tcp_sock.setblocking(0)
            self._tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._tcp_sock.bind(self.addr)
            self._tcp_sock.listen(128)  # max 128 clients
        except socket.error:
            self.logger.warning('Unable to bind TCP socket at addr={}'.format(self.addr))
        else:
            self.logger.info("Listening on TCP addr={}".format(self.addr))

        return self._tcp_sock

    @property
    def udp_sock(self):
        if self._udp_sock:
            return self._udp_sock

        try:
            self._udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self._udp_sock.setblocking(0)
            self._udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._udp_sock.bind(self.addr)
        except socket.error:
            self.logger.warning('Unable to bind UDP socket at addr={}'.format(self.addr))
        else:
            self.logger.info("Listening on UDP addr={}".format(self.addr))

        return self._udp_sock

    def close_tcp(self):
        self._tcp_sock = None

    def close_udp(self):
        self._udp_sock = None

    def connect(self, info):
        conn, addr = info
        self.logger.info('%s, %s connected' % (conn, addr))

        self.connections[conn] = addr

    def disconnect(self, connection):
        addr = self.connections.pop(connection)
        self.logger.info('%s, %s disconnected' % (connection, addr))

    def fileno(self):
        return self.tcp_sock.fileno()

    def handle_message(self, client, buf):
        self.logger.info('client={}, buf={}'.format(client, buf))

    def loop(self, select_timeout=1.0):
        handled = None

        # check UDP
        try:
            buf, client = self.udp_sock.recvfrom(1024)
        except socket.error, exc:
            # (11, 'Resource temporarily unavailable')
            # [Errno 35] Resource temporarily unavailable
            error_number = exc[0]
            if error_number not in (11, 35):
                self.logger.exception(exc)
                self.logger.warning('got socket error_number={}'.format(error_number))

                self.close_udp()
        else:
            if buf:
                self.handle_message(client, buf)
                handled = True

        try:
            self.connect(self.tcp_sock.accept())
        except socket.error, exc:
            # (11, 'Resource temporarily unavailable')
            # [Errno 35] Resource temporarily unavailable
            error_number = exc[0]
            if error_number not in (11, 35):
                self.logger.exception(exc)
                self.logger.warning('got socket error_number={}'.format(error_number))

                self.close_tcp()

        ready, _, _ = select(self.connections, [], [], select_timeout)

        for client in ready:
            try:
                buf = client.recv(BUFSIZE)
                if buf == '':
                    self.disconnect(client)
                    continue
            except socket.error, exc:
                # [Errno 54] Connection reset by peer
                # [Errno 104] Connection reset by peer -- raspbian
                error_number = exc[0]
                if error_number not in (54, 104):
                    self.logger.exception(exc)
                    self.logger.warning('got socket error_number={}'.format(error_number))

                self.disconnect(client)
                continue

            self.handle_message(client, buf)
            handled = True

        if self.message_handler:
            self.message_handler.loop()

        return handled

    def quit(self):
        self.tcp_sock.close()
        self.udp_sock.close()

        # prevent getting exception where dictionary changes while looping
        connections = self.connections.keys()
        for connection in connections:
            self.disconnect(connection)

    def run(self):
        while self._run:
            self.loop()


class MessengerServer(BaseServer):
    def handle_message(self, client, buf):
        for line in buf.splitlines():
            if self.message_handler:
                self.message_handler.handle_message(client, self.connections[client], line)
            else:
                message = 'client={}, line={}'.format(self.connections[client], line)
                self.logger.info(message)

                try:
                    data = json.loads(line)
                except ValueError:
                    self.logger.error('unable to parse line={}'.format(line))
                else:
                    data['meta'] = {
                        'client': self.connections[client],
                    }

                    client.send(json.dumps(data))


class ConsoleMessengerServer(MessengerServer):
    port = 9628
    host = 'localhost'

    def __init__(self, host=None, port=None, message_handler=None):
        host = host or self.host
        port = port or self.port

        super(ConsoleMessengerServer, self).__init__(host, port, message_handler=message_handler)

    @property
    def connector(self):
        return get_connector()

    def handle_message(self, client, buf):
        message = parse_buf(buf)

        try:
            self.connector.send(message)
        except NotConnected:
            self.logger.warning('message={} dropped, not connected to upstream'.format(message))

    def loop(self, select_timeout=0.01):
        return super(ConsoleMessengerServer, self).loop(select_timeout=select_timeout)


if __name__ == '__main__':
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument('--loglevel', default='info')
    parser.add_argument('--host', default='localhost')
    parser.add_argument('--port', type=int, default=9628)

    args = parser.parse_args()

    loglevel = getattr(logging, args.loglevel.upper())
    logging.basicConfig(stream=sys.stdout, level=loglevel)

    server = BaseServer(args.host, args.port)
    try:
        server.run()
    except KeyboardInterrupt:
        server.quit()



"""
This is a stub package that allows the compat import mechanism injected in
zymbit/__init__.py

The importer will check the system this is running on, e.g. arduino,
raspberrypi, etc. and import modules from that package
"""



import json
import logging
import socket
import sys
import time

from select import select

import websocket

from backoff import Backoff

from zymbit.envelope import get_envelope


ACK_TIMEOUT = 5 # 5 seconds to get an ack
BUFSIZE = 4096
MAX_BACKOFF = 300 # retry connection once a minute


class MessengerClient(object):
    def __init__(self, host=None, port=None):
        self.addr = (host, port)

        self.sock = None

        # keep track of the last message sent in case the socket goes boom.  in
        # that case, requeue the last message to go out again.
        self.last_message = None

        # messages that have not been sent
        self.outbound_queue = []

        self.inbound_queue = []

    @property
    def logger(self):
        name = '{}.{}'.format(__name__, self.__class__.__name__)

        return logging.getLogger(name)

    @Backoff()
    def _check_connection(self):
        if self.sock:
            return True

        try:
            self.connect()
        except socket.error, exc:
            self.sock = None

            if exc[0] not in (61, 111):
                raise

            return False

        return True

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect(self.addr)

        self.logger.debug('connected to %s' % (self.addr,))

    def fileno(self):
        return self.sock.fileno()

    def loop(self):
        """
        Process the outbound message queue
        """

        if not self._check_connection():
            return False

        ready, _, _ = select([self.sock], [], [], 0.01)
        if ready:
            try:
                buf = self._recv()

                # when an empty buffer is received, the connection is dead.
                if buf in (None, ''):
                    self.sock = None
                    return False

                self.logger.debug('buffer: %r' % (buf,))
                for line in buf.splitlines():
                    # skip empty lines
                    if not line:
                        continue

                    self.inbound_queue.append(json.loads(line))
            except socket.error, exc:
                # [Errno 54] Connection reset by peer
                # socket.error: (104, 'Connection reset by peer')
                if exc[0] not in (54, 104):
                    raise

                self.sock = None

                return False

        while self.outbound_queue:
            try:
                envelope = get_envelope(*self.outbound_queue[0])
                self._send('{}\n'.format(envelope))
            except socket.error, exc:
                # [Errno 32] Broken pipe
                if exc[0] != 32:
                    raise

                self.sock.close()
                self.sock = None

                break
            else:
                # remove from the queue
                self.outbound_queue.pop(0)

    def queue(self, action, params):
        self.outbound_queue.append((action, params))

    def _recv(self):
        return self.sock.recv(BUFSIZE)

    def recv(self):
        if self.inbound_queue:
            return self.inbound_queue.pop(0)

    def _send(self, envelope):
        self.sock.send(envelope)

    def send(self, action, params):
        print 'adding to queue, action={}, params={}'.format(action, params)
        self.queue(action, params)


class WebsocketMessengerClient(MessengerClient):
    def __init__(self, url):
        super(WebsocketMessengerClient, self).__init__()

        self.client = None
        self.sock = None

        self.url = url

    def _check_incoming(self):
        if not self.sock:
            return

        try:
            message = self.client.recv()

            self.inbound_queue.append(message)

            return True
        except websocket.WebSocketException, exc:
            return False
        except socket.error, exc:
            #socket.error: [Errno 35] Resource temporarily unavailable
            #socket.error: [Errno 54] Connection reset by peer
            if exc[0] not in (35, 54):
                raise

    def connect(self):
        print 'try to connect to websocket'
        try:
            self.client = websocket.create_connection(self.url)
            self.client.settimeout(0)

            self.sock = self.client.sock
        except socket.error, exc:
            self.client = None
            self.sock = None

            raise

    def fileno(self):
        return self.sock.fileno()

    def _recv(self):
        try:
            return self.client.recv()
        except websocket.WebSocketException, exc: # Not a valid frame None
            self.client = None
            self.sock = None
        except socket.error, exc:
            # [Errno 35] Resource temporarily unavailable
            if exc[0] != 35:
                raise

            self.client = None
            self.sock = None

    def _send(self, envelope):
        buf = json.dumps(envelope)
        self.client.send(buf)




from __builtin__ import classmethod
import logging
import requests
import sys

from zymbit.commands import Command

from zymbit.config import get_config, save_config


class PasswordCommand(Command):
    command_name = 'password'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        config = get_config()

        default_auth = config['auth']['default_auth']
        auth = config['auth'][default_auth]

        print auth['password']



from __future__ import absolute_import

import inspect
import os
import shutil

from zymbit.util import run_command

from . import config
from . import util


def call_distro_fn(*args, **kwargs):
    distro = util.get_distro()

    last_frame = inspect.currentframe().f_back
    fn_name = inspect.getframeinfo(last_frame).function
    distro_fn_name = '{}_{}'.format(fn_name, distro)

    return globals()[distro_fn_name](*args, **kwargs)


def enable_init_script():
    return call_distro_fn()


def enable_init_script_arch():
    command = ('/usr/bin/systemctl', 'enable', 'zymbit')

    return run_command(command)


def enable_init_script_raspbian():
    os.chmod(config.INIT_SCRIPT_PATH, 0755)

    script_dir = os.path.dirname(config.INIT_SCRIPT_PATH)

    command = ('/usr/sbin/update-service', '--add', script_dir)

    return run_command(command)


def setup_metrics():
    return call_distro_fn()


def setup_metrics_arch():
    return 0  # the file is already in the correct location


def setup_metrics_raspbian():
    dest = '/etc/collectd/collectd.conf'

    dest_dir = os.path.dirname(dest)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    shutil.move('/etc/collectd.conf', dest)

    return 0


def start_metrics():
    return call_distro_fn()


def start_metrics_arch():
    # restart because the collectd service is presumably already installed
    # and we're restarting it so it loads our config.
    command = ('/usr/bin/systemctl', 'restart', 'collectd')

    return run_command(command)


def start_metrics_raspbian():
    # restart because the collectd service is presumably already installed
    # and we're restarting it so it loads our config.
    command = ('/etc/init.d/collectd', 'restart')

    return run_command(command)


def start_service():
    return call_distro_fn()


def start_service_arch():
    command = ('/usr/bin/systemctl', 'start', 'zymbit')

    return run_command(command)


def start_service_raspbian():
    script_dir = os.path.dirname(config.INIT_SCRIPT_PATH)
    command = ('/usr/bin/supervise', script_dir)

    return run_command(command)



import logging

from zymbit.commands import Command
from zymbit.util import get_provisioning_token


class ProvisioningTokenCommand(Command):
    command_name = 'provisioning_token'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        print get_provisioning_token()



import logging

from tabulate import tabulate

from zymbit.commands import Command

from zymbit.util import ZymbitApi


class DevicesCommand(Command):
    command_name = 'devices'
    api_endpoint = '/zymbots'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        api = ZymbitApi()
        response = api.request('get', self.api_endpoint)

        data = response.json()

        headers = ['hostname', 'client_id', 'uuid']
        table = [[x['hostname'], x['client_id'], x['uuid']] for x in data['results']]

        print tabulate(table, headers=headers)



import logging
import sys

import requests

from zymbit.commands import Command
from zymbit.config import get_config, save_config

API_PREFIX = '/api/v1'
WEBSOCKET_URL = '/websocket_url'


class CloudCommand(Command):
    command_name = 'cloud'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(CloudCommand, cls).setup_parser(parser)

        command_parser.add_argument('--insecure', action='store_true')
        command_parser.add_argument('api_uri')

        return command_parser

    def run(self):
        api_uri = self.args.api_uri
        verify = not self.args.insecure

        self.logger.debug('api_uri={}, verify={}'.format(api_uri, verify))

        session = requests.Session()

        api_url = '{}{}'.format(api_uri, API_PREFIX)
        websocket_url_endpoint = '{}{}'.format(api_url, WEBSOCKET_URL)

        response = session.get(websocket_url_endpoint, verify=verify)
        try:
            response.raise_for_status()
        except requests.HTTPError, exc:
            status_code = response.status_code
            sys.exit('got status_code={} when getting websocket URL'.format(status_code))

        websocket_url = response.json()
        
        config = get_config()
        cloud_config = config.get('cloud', {})
        cloud_config.update({
            'check_hostname': verify,
            'api_url': api_url,
            'websocket_url': websocket_url,
        })

        config['cloud'] = cloud_config
        save_config(config)



from zymbit.config import *

INIT_SCRIPT_PATH = '/opt/zymbit/run'
FILES_ROOT = '/usr/local/files/raspberrypi'
KNOWN_HOSTS = '/etc/ssh/ssh_known_hosts'
ZYMBIT_KNOWN_HOSTS = '/etc/ssh/zymbit_known_hosts'



from subprocess import PIPE, Popen

from zymbit.util import *


def create_ssh_key(identity_file):
    dirname = os.path.dirname(identity_file)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    command = ('/usr/bin/dropbearkey', '-t', 'rsa', '-s', '2048', '-f', identity_file)
    return run_command(command)


def get_distro():
    if os.path.exists('/etc/linino'):
        return 'linino'


def get_pubkey(identity_file):
    command = ('/usr/bin/dropbearkey', '-y', '-f', identity_file)
    proc = Popen(command, stdout=PIPE)
    output = proc.communicate()[0]

    for line in output.splitlines():
        if line.startswith('ssh-rsa'):
            return line



from zymbit.commands import Command
from zymbit.provision import Provision


class ProvisionCommand(Command):
    command_name = 'provision'

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(ProvisionCommand, cls).setup_parser(parser)

        command_parser.add_argument('--api-url')
        command_parser.add_argument('--websocket-url')
        command_parser.add_argument('--ssl-ignore-hostname', action='store_true')
        command_parser.add_argument('--config-only', action='store_true', help='only write config and exit')
        command_parser.add_argument('--clean', action='store_true', help='remove existing config')
        command_parser.add_argument('token')

        return command_parser

    def run(self):
        token = self.args.token
        api_url = self.args.api_url
        websocket_url = self.args.websocket_url
        check_hostname = not self.args.ssl_ignore_hostname
        config_only = self.args.config_only
        clean = self.args.clean

        provision = Provision(
            token,
            api_url=api_url,
            websocket_url=websocket_url,
            check_hostname=check_hostname,
            config_only=config_only,
            clean=clean
        )
        provision.run()



import logging
import socket
import time
from zymbit.exceptions import ConnectionError

from zymbit.linux.console import Console

from subprocess import PIPE, STDOUT, Popen

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6571


class ArduinoConsole(Console):
    @property
    def logger(self):
        return logging.getLogger(__name__)

    def connected(self, host_info):
        super(ArduinoConsole, self).connected(host_info)

        self.logger.debug('connected to host_info={}'.format(host_info))

    def _restart_bridge(self):
        """
        Checks to see if the bridge process is running
        :return: boolean - whether the bridge was restarted
        """
        running = self._is_bridge_running()
        if not running:
            self.logger.warning('bridge not running, resetting mcu')
            status = Popen('/usr/bin/reset-mcu').wait()
            self.logger.info('reset-mcu status={}'.format(status))

            for i in range(10):  # wait for bridge to show up
                running = self._is_bridge_running()
                if running:
                    break

                time.sleep(1.0)
            else:
                self.logger.warning('bridge still not found in process listing')

            if running:  # let the bridge settle for a couple seconds
                time.sleep(2.0)

            return running

        return False

    def handle_socket_exception(self, socket_exc):
        """
        Handle the socket exception
        :param socket_exc: Exception
        """
        super(ArduinoConsole, self).handle_socket_exception(socket_exc)

        # try to restart the bridge on arduinos and if successful, try reopening the socket
        if self._restart_bridge():
            return self.socket

    def _is_bridge_running(self):
        """
        Check to see if bridge is running
        :return: Boolean
        """
        command = 'ps'

        proc = Popen(command, stdout=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate('')

        running = 'python -u bridge.py' in stdout

        return running


class ArduinoConsoleTask(ArduinoConsole):
    """
    Subclass that relays all arduino messages to the cloud messenger
    """
    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT):
        super(ArduinoConsoleTask, self).__init__(host, port)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.addr = ('localhost', 9628)

    @property
    def logger(self):
        name = '{}.{}'.format(__name__, self.__class__.__name__)

        return logging.getLogger(name)

    def loop(self):
        # TODO: add a buffer to store messages that cannot be sent
        try:
            buf = self.recv(0.1)
        except ConnectionError, exc:
            self.handle_socket_exception(exc)
        else:
            if not buf:
                return

            for line in buf.splitlines():
                self.logger.debug('sending line={} to addr={}'.format(line, self.addr))
                try:
                    self.sock.sendto(line, self.addr)
                except socket.error, exc:
                    self.logger.exception(exc)
                    self.logger.error('unable to send buf={}'.format(buf))

                    self.handle_socket_exception(exc)



import logging.config
import os
import sys

from zymbit.util import is_interactive

LOGLEVEL = os.environ.get('LOGLEVEL', 'info')
LOGFILE = os.environ.get('LOGFILE')


def config_logging(logfile=None, loglevel=None):
    logging.captureWarnings(True)

    logging.config.dictConfig(get_logging_config(logfile, loglevel))


def get_logging_config(logfile=None, loglevel=None):
    logfile = logfile or LOGFILE
    loglevel = (loglevel or LOGLEVEL).upper()

    syslog_device_path = '/dev/log'
    if sys.platform == 'darwin':
        syslog_device_path = '/var/run/syslog'

    logging_config = {
        'version': 1,
        'disable_existing_loggers': True,

        'formatters': {
            'default': {
                'format': '%(levelname)s %(asctime)s %(name)s %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
        },

        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': loglevel,
                'stream': 'ext://sys.stdout',
                'formatter': 'default',
            },
            'syslog': {
                'class': 'logging.handlers.SysLogHandler',
                'level': loglevel,
                'address': syslog_device_path,
                'formatter': 'default',
            },
            'null': {
                'class': 'logging.NullHandler',
                'level': 'DEBUG',
            },
        },

        'loggers': {
            '': {
                'handlers': [],
                'propagate': True,
                'level': loglevel,
            },
            'filer': {
                'handlers': [],
                'propagate': False,
                'level': loglevel,
            },
            'requests': {
                'handlers': ['null'],
                'propagate': False,
                'level': 'ERROR',
            },
        },
    }

    if (is_interactive() and not logfile) or logfile == '-':  # log to the console when - is given as the logfile
        logging_config['loggers']['']['handlers'].append('console')
        logging_config['loggers']['filer']['handlers'].append('console')

        logging_config['handlers']['console'].update({
            'stream': 'ext://sys.stderr',
        })
    elif logfile:  # any other logfile setup a rotating file handler
        logging_config['handlers'].update({
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': 'INFO',
                'formatter': 'default',
                'filename': logfile,
                'maxBytes': 1024 * 100,
                'backupCount': 2,
            },
        })

        logging_config['loggers']['']['handlers'].append('file')
        logging_config['loggers']['filer']['handlers'].append('file')
    else:  # log to syslog by default
        logging_config['loggers']['']['handlers'].append('syslog')
        logging_config['loggers']['filer']['handlers'].append('syslog')

    return logging_config



import getpass
import logging
import requests
import sys

from zymbit.commands import Command

from zymbit.config import get_config, save_auth
from zymbit.util import validate_email


class LoginCommand(Command):
    command_name = 'login'
    api_endpoint = '/members/login'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(LoginCommand, cls).setup_parser(parser)

        # use type kwarg to validate email address
        # http://stackoverflow.com/a/14117511/703144
        command_parser.add_argument('email', type=validate_email)

        return command_parser

    def run(self):
        password = getpass.getpass()

        params = {
            'username': self.args.email,
            'password': password,
        }

        config = get_config()
        api_url = config.lookup('cloud.api_url')
        verify = config.lookup('cloud.check_hostname')

        url = '{}{}'.format(api_url, self.api_endpoint)

        try:
            response = requests.post(url, data=params, verify=verify)
        except requests.ConnectionError, exc:
            self.logger.error('Unable to connect to url={}'.format(url))

            self.logger.exception(exc)
        else:
            try:
                response.raise_for_status()
            except requests.HTTPError:
                data = response.json()
                sys.exit(data)

            save_auth(response.json(), password)



from __future__ import absolute_import

import logging
import socket
import time

from zymbit.exceptions import NotConnected
from zymbit.websocket import WebSocketConnectionClosedException

from .client import Client


class Stream(Client):
    def __init__(self, subscriptions):
        self.subscriptions = subscriptions

        super(Stream, self).__init__(connected_callback=self.online)

    @property
    def logger(self):
        name = '{}.{}'.format(__name__, self.__class__.__name__)
        return logging.getLogger(name)

    def handle(self, envelope):
        pass

    def loop(self):
        try:
            envelope = self.recv()
        except NotConnected:
            time.sleep(1.0)
        else:
            if envelope is None:
                time.sleep(1.0)
                return

            self.handle(envelope)

    def online(self):
        """
        Called when websocket connection is initialized
        """
        self.logger.debug('connected, subscribing ...')
        for routing_key in self.subscriptions:
            params = {'routing_key': routing_key}
            self.logger.debug('subscribe params={}'.format(params))
            self.send('subscribe', params)

    def run(self):
        while True:
            self.loop()



import json
import logging

from zymbit.client import Client
from zymbit.commands import Command


class PublishCommand(Command):
    command_name = 'publish'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(PublishCommand, cls).setup_parser(parser)

        command_parser.add_argument('--route', default='data')
        command_parser.add_argument('key')
        command_parser.add_argument('value', help='JSON encoded, e.g. "3" is an int \'"3"\' is a string')

        return command_parser

    def run(self):
        client = Client()
        value = json.loads(self.args.value)

        client.send(self.args.route, {'key': self.args.key, 'value': value})

        none_count = 0
        while True:
            data = client.recv()
            if data is None:
                none_count += 1
                if none_count > 10:
                    break
            else:
                print data



import logging

from zymbit.commands import Command
from zymbit.config import get_config, save_config


class LogoutCommand(Command):
    command_name = 'logout'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def run(self):
        config = get_config()

        auth_config = config.get('auth')

        if not auth_config:
            return

        auth_config.pop('default_auth', None)

        keys = auth_config.keys()
        for key in keys:
            if '@' in key and '.' in key:
                auth_config.pop(key)

        save_config(config)



from __builtin__ import classmethod
import logging
import sys

import requests

from zymbit.commands import Command
from zymbit.util import ZymbitApi


class ConfirmCommand(Command):
    command_name = 'confirm'
    api_endpoint = '/registration/{}/confirm'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(ConfirmCommand, cls).setup_parser(parser)

        command_parser.add_argument('confirmation_code')

        return command_parser

    def run(self):
        api = ZymbitApi()
        endpoint = self.api_endpoint.format(self.args.confirmation_code)
        try:
            response = api.request('post', endpoint)
        except requests.ConnectionError, exc:
            sys.exit(exc)
        except requests.HTTPError:
            access_okay = False

            error_response = api.response
            if api.response.status_code == 404:
                response = api.request('get', '/zymbots')
                if response.status_code == 200:
                    access_okay = True
                else:
                    sys.exit('confirmation code not found, are you already confirmed?')

            if not access_okay:
                data = error_response.json()
                sys.exit(data)

        print 'Your email address is confirmed'



from __future__ import absolute_import

import importlib
import json
import logging
import socket

from zymbit import config, statemachine, util
from zymbit.client import Client, register
from zymbit.clientid import get_client_id
from zymbit.compat import config, util
from zymbit.exceptions import NotConnected, ProvisionError, TasksFailed

CONNECTION_REFUSED = 61


def get_connector():
    return Connector.instance()


class Connector(statemachine.StateMachine):
    _instance = None

    def __init__(self):
        super(Connector, self).__init__()

        self.client = None
        self.system = util.get_system()

        # tasks that should be run on loop
        self.tasks = {}
        self.tasks_config = {}

    @classmethod
    def instance(cls):
        if cls._instance:
            return cls._instance

        cls._instance = cls()

        return cls._instance

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def connect(self):
        if not self.client:
            self.client = Client()

        # don't reconnect an already-connected client
        connected = self.client.is_connected
        if connected:
            return True

        try:
            self.client.connect()
        except NotConnected:
            pass
        except socket.error, exc:
            if exc[0] not in (111,):
                raise
        else:
            self.logger.info('connected')
            self.load_tasks()

            return True

        # process tasks if not connected
        self.process_tasks()

    def connected(self):
        try:
            line = self.client.recv()
        except socket.error, exc:
            if exc[0] not in (110, 111):
                raise

            self.client.close()
            raise NotConnected
        else:
            if line:
                self.process_message(line)

        processed = self.process_tasks()
        # if tasks are processed, don't sleep
        if processed:
            self.loop_sleep_time = 0.01
        else:
            self.loop_sleep_time = 1.0

        # self.logger.debug('processed={}, loop_sleep_time={}'.format(processed, self.loop_sleep_time))

    def get_auth(self):
        _config = config.get_config()
        client_id = get_client_id()

        return _config.get('auth', {}).get(client_id)

    def on_config(self, data):
        _config = data['params']
        config.save_config(_config)

    def process_message(self, message):
        try:
            data = json.loads(message)
        except ValueError:
            pass
        else:
            action = data['action']
            action_fn = getattr(self, 'on_{}'.format(action), None)
            if action_fn:
                action_fn(data)
                return

        self.logger.info('message={}'.format(message))

    def load_task(self, name, task_config):
        module_path, class_name = task_config['class_path'].rsplit('.', 1)

        task_module = importlib.import_module(module_path)
        task_class = getattr(task_module, class_name)

        task = task_class(**task_config.get('params', {}))

        self.logger.debug('load task name={}, task_config={}, task_class={}, task={}'.format(
            name, task_config, task_class, task)
        )

        self.tasks[name] = task

    def load_tasks(self):
        _config = config.get_config()
        self.tasks_config = _config.get('tasks', {})

        self.logger.debug('load_tasks, tasks_config={}'.format(self.tasks_config))
        for name, task_config in self.tasks_config.items():
            # if a task is already loaded, skip
            if name in self.tasks:
                self.logger.debug('task name={} already loaded'.format(name))
                continue

            try:
                self.load_task(name, task_config)
                self.logger.info('loaded task name={}'.format(name))
            except Exception, exc:
                self.logger.exception(exc)
                self.logger.error('task name={} raised exception when trying to load it'.format(name))

        return True

    def process_tasks(self):
        processed = None

        reload_tasks = False
        task_names = self.tasks.keys()
        for task_name in task_names:
            task = self.tasks[task_name]

            try:
                # self.logger.debug('task task_name={}, task={}'.format(task_name, task))
                if task.loop():
                    processed = True
            except Exception, exc:
                self.logger.exception(exc)
                self.logger.error('task_name={} raised exception above'.format(task_name))

                self.quit_task(task_name, task)
                reload_tasks = True

        if reload_tasks:
            raise TasksFailed

        return processed

    def quit(self):
        for name, task in self.tasks.items():
            self.quit_task(name, task)

    def quit_task(self, name, task):
        try:
            task.quit()
        except Exception, exc2:
            self.logger.exception(exc2)
            self.logger.error('task name={} raised exception above trying to quit'.format(name))

        self.tasks.pop(name, None)  # don't barf if the task is not in the tasks list

    def reload_tasks(self):
        return self.load_tasks()

    def register(self):
        client_id = get_client_id()
        auth = self.get_auth()
        if auth:
            return True

        try:
            device_config = register()
            config.save_config(device_config)

            return True
        except ProvisionError:
            self.logger.debug(
                'waiting for provisioning token, no auth for client_id={}'.format(client_id))

    def run(self):
        while self._run:
            try:
                super(Connector, self).run()
            except (NotConnected, TasksFailed):
                pass

    def send(self, envelope):
        self.client.send_raw(envelope)

    def subscribe(self):
        return True

    # now that the methods are defined, use the function definitions themselves
    # to setup the state machine transitions
    transitions = {
        statemachine.StateMachine.start: {
            True: load_tasks,
        },
        load_tasks: {
            True: register,
        },
        register: {
            True: connect,
        },
        connect: {
            True: subscribe,
            TasksFailed: reload_tasks,
        },
        subscribe: {
            True: connected,
            NotConnected: connect,
        },
        connected: {
            NotConnected: connect,
            TasksFailed: reload_tasks,
        },
        reload_tasks: {
            True: connect,
        },
    }



import json
import mock

from unittest import TestCase
from zymbit.messenger.server import ConsoleMessengerServer


class ConsoleMessengerServerTestCase(TestCase):
    @mock.patch('zymbit.messenger.server.get_connector')
    @mock.patch('zymbit.messenger.server.socket')
    def test_int_json(self, socket_mock, get_connector_mock):
        client = mock.Mock()

        server = ConsoleMessengerServer()
        server.connect((client, ('host', 12345)))

        server.handle_message(client, '3\n')

        send = get_connector_mock.return_value.send
        self.assertEqual(1, send.call_count)

        envelope = send.call_args[0][0]
        data = json.loads(envelope)

        self.assertEqual({'key': 'sensor', 'value': 3}, data['params'])



#!/usr/bin/env python
import os

from distutils.core import setup

SCRIPT_DIR = os.path.dirname(__file__)
if not SCRIPT_DIR:
        SCRIPT_DIR = os.getcwd()


# put together list of requirements to install
install_requires = []
with open(os.path.join(SCRIPT_DIR, 'requirements.txt')) as fh:
    for line in fh.readlines():
        if line.startswith('-'):
            continue

        install_requires.append(line.strip())

data_files = [(dirpath, [os.path.join(dirpath, x) for x in filenames]) for dirpath, dirnames, filenames in os.walk('files') if filenames]

setup(name='zymbit',
      version='0.5.3',
      description='Zymbit cloud library',
      author='Roberto Aguilar',
      author_email='roberto@zymbit.com',
      packages=['zymbit', 'zymbit.arduino', 'zymbit.commands', 'zymbit.compat', 'zymbit.darwin', 'zymbit.linux', 'zymbit.messenger', 'zymbit.raspberrypi'],
      scripts=['scripts/zymbit'],
      data_files=data_files,
      long_description=open('README').read(),
      url='http://datavytals.com/',
      license='LICENSE',
      install_requires=install_requires,
)



import logging
import random
import requests
import sys

from zymbit.commands import Command

from zymbit.config import get_config, save_auth
from zymbit.util import validate_email


def random_string(length=20):
    chars = (
        'abcdefghijklmnopqrstuvwxyz'
        'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        '0123456789'
        '!@#$%^*|-+='
    )

    selected = []
    for _ in range(20):
        selected.append(random.choice(chars))

    return ''.join(selected)


class SignupCommand(Command):
    command_name = 'signup'
    api_endpoint = '/registration'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(SignupCommand, cls).setup_parser(parser)

        # use type kwarg to validate email address
        # http://stackoverflow.com/a/14117511/703144
        command_parser.add_argument('email', type=validate_email)

        return command_parser

    def run(self):
        # randomly generate a password
        password = random_string()

        params = {
            'email': self.args.email,
            'password': password,
        }

        config = get_config()
        api_url = config['cloud']['api_url']
        verify = config['cloud'].get('check_hostname', True)

        url = '{}{}'.format(api_url, self.api_endpoint)

        try:
            response = requests.post(url, data=params, verify=verify)
        except requests.ConnectionError, exc:
            self.logger.error('Unable to connect to url={}'.format(url))

            self.logger.exception(exc)
        else:
            try:
                response.raise_for_status()
            except requests.HTTPError:
                data = response.json()
                sys.exit(data)

            # save the login information
            auth_data = response.json()

            try:
                save_auth(auth_data, password)
            except requests.HTTPError:
                data = response.json()
                sys.exit(data)

            print 'You are signed up; check your email at {} for the verification key'.format(self.args.email)



from zymbit.config import *

FILES_ROOT = '/usr/files/arduino'
KNOWN_HOSTS = '/.ssh/known_hosts'

TUNNEL_SSH_KEY_PATH = os.path.expanduser(os.environ.get('TUNNEL_SSH_KEY_PATH', '/.ssh/id_tunnel'))

ZYMBIT_KNOWN_HOSTS = '/etc/dropbear/zymbit_known_hosts'



import logging

from zymbit.commands import Command
from zymbit.stream import Stream


class EchoStream(Stream):
    def handle(self, envelope):
        print envelope


class SubscribeCommand(Command):
    command_name = 'subscribe'

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @classmethod
    def setup_parser(cls, parser):
        command_parser = super(SubscribeCommand, cls).setup_parser(parser)

        command_parser.add_argument('subscriptions', nargs='+')

        return command_parser

    def run(self):
        EchoStream(self.args.subscriptions).run()



import datetime
import logging
import os
import select
import shutil

from subprocess import PIPE, STDOUT, Popen

from zymbit.compat import config, util
from zymbit.exceptions import TunnelError
from zymbit.timeutil import now

SSH = os.environ.get('SSH', '/usr/bin/ssh')
MAX_BACKOFF = 300  # 5 minutes

# how long a connection must be in place in order to consider it okay to clear backoff
CONNECTION_OK_DELTA = datetime.timedelta(seconds=30)


class Tunnel(object):
    defaults = {
        'user': None,
        'host': None,
        'identity_file': config.TUNNEL_SSH_KEY_PATH,
        'ssh_port': None,
        'port': 22,
    }

    def __init__(self, **kwargs):
        _config = self.defaults.copy()
        _config.update(config.get_config().get('tasks', {}).get('tunnel', {}).get('config', {}))
        _config.update(dict([item for item in kwargs.items() if item[1]]))

        unconfigured = [key for key, val in _config.items() if val is None]
        if unconfigured:
            raise TunnelError('{} cannot be unset'.format(unconfigured))

        self.user = _config['user']
        self.host = _config['host']
        self.port = str(_config['port'])
        self.identity_file = _config['identity_file']
        self.ssh_port = str(_config['ssh_port'])

        self.connection = None
        self.connected_at = None

        self.backoff = 0
        self.backoff_until = None

        self.connect()

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def close(self):
        if self.connection is None:
            return

        try:
            self.connection.kill()
        except OSError, exc:
            if exc[0] != 3:
                raise

        self.connection.wait()

        self.connection = None

    def check(self):
        if not self.connection:
            self.connect()
            if self.connection is None:
                self.logger.debug('no connection')
                return

        # this is the main throttling sleep here, keep it at around 1 second
        stdout = self.connection.stdout
        r, _, _ = select.select([stdout], [], [], 0.01)
        for stream in r:
            line = stream.readline().strip()
            if line:
                self.logger.info('line={}'.format(line))

        status = self.connection.poll()
        if status is None:
            # connection is established, see if backoff needs to be reset
            if self.backoff and now() > self.connected_at + CONNECTION_OK_DELTA:
                self.logger.debug('connection ok, resetting backoff')
                self.backoff = 0
                self.backoff_until = None

            return

        # a connection had been established, but has ended, need to reconnect
        self.logger.warning('closing connection after status={}'.format(status))
        self.close()

        # set the backoff time
        self.backoff += 1
        seconds = min(2**self.backoff, MAX_BACKOFF)
        self.backoff_until = now() + datetime.timedelta(seconds=seconds)

        self.logger.debug('backoff={}, seconds={}, backoff_until={}'.format(
            self.backoff, seconds, self.backoff_until))

    def check_host_key(self):
        zymbit_known_hosts = config.ZYMBIT_KNOWN_HOSTS
        known_hosts = config.KNOWN_HOSTS
        if not os.path.exists(known_hosts):
            shutil.copyfile(zymbit_known_hosts, known_hosts)
            return True

        with open(zymbit_known_hosts, 'rb') as zbkh:
            while True:
                pubkey = zbkh.readline()
                if pubkey == '':  # end of file
                    break

                pubkey = pubkey.strip()
                if pubkey == '':  # skip empty lines
                    continue

                known_hosts_temp = '{}.tmp'.format(known_hosts)

                try:
                    line_found = False
                    with open(known_hosts_temp, 'wb') as new_hosts:
                        with open(known_hosts, 'rb') as old_hosts:
                            line = old_hosts.readline().strip()

                            if line == pubkey:
                                line_found = True

                            new_hosts.write('{}\n'.format(line))

                        if not line_found:
                            new_hosts.write('{}\n'.format(pubkey))

                    # overwrite the temporary file
                    shutil.move(known_hosts_temp, known_hosts)
                finally:
                    if os.path.exists(known_hosts_temp):
                        os.remove(known_hosts_temp)

    def connect(self):
        if self.backoff_until and self.backoff_until > now():
            self.logger.debug('not connecting, backoff_until={}'.format(self.backoff_until))
            return

        self.create_ssh_key()
        self.check_host_key()

        self.connection = Popen((
            SSH, '-N', '-I', '60', '-R', self.tunnel_conf,
            '-i', self.identity_file, '-p', self.port, self.remote_server
        ), stdout=PIPE, stderr=STDOUT)

        self.connected_at = now()

        return self.connection

    def create_ssh_key(self):
        _config = config.get_config()
        tunnel_config = _config.setdefault('tunnel', {})
        identity_file = config.TUNNEL_SSH_KEY_PATH
        if not os.path.exists(identity_file):
            util.create_ssh_key(identity_file)
            self.logger.info('created ssh key at identity_file={}'.format(identity_file))

        # check to see if config should be written
        if not tunnel_config.get('host'):
            _config['tunnel'].update({
                'user': 'tunnel',
                'host': 'tunnel.zymbit.com',
            })
            config.save_config(_config)

        return True

    def loop(self):
        return self.check()

    def reconnect(self):
        self.close()
        self.connect()

    @property
    def remote_server(self):
        return '{}@{}'.format(self.user, self.host)

    @property
    def tunnel_conf(self):
        return '{}:localhost:22'.format(self.ssh_port)



from __future__ import absolute_import

# importing on the yun requires the try block, importing on a desktop
# system works with the except block (?!)
try:
    from websocket import websocket

    from websocket.websocket import WebSocketConnectionClosedException
except ImportError:
    import websocket

    from websocket import WebSocketConnectionClosedException
